
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

enum {
	// _xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 4_000_000,

	MOSI = 16,
	CLK = 17,
	CS = 18,
	MISO = 19,

	CPOL = 0,   // 0 or 1 only, used in binary equation
        CPHA = 0,   // 0 or 1 only, used in binary equation
};


/*
 *      _____      Mode 0 : CPOL=0  CPHA=0      ____
 *   CS      |_________________________________|
 *              _   _   _   _   _   _   _   _
 *  CLK _______| |_| |_| |_| |_| |_| |_| |_| |______
 *            ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI -----X___X___X___X___X___X___X___X___XXX----
 * MISO        7   6   5   4   3   2   1   0
 * 
 * 
 *      _____      Mode 1 : CPOL=0  CPHA=1      ____
 *   CS      |_________________________________|
 *              _   _   _   _   _   _   _   _
 *  CLK _______| |_| |_| |_| |_| |_| |_| |_| |______
 *              ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI -----XXX___X___X___X___X___X___X___X___X----
 * MISO          7   6   5   4   3   2   1   0
 * 
 * 
 *      _____      Mode 2 : CPOL=1  CPHA=0      ____
 *   CS      |_________________________________|
 *      _______   _   _   _   _   _   _   _   ______
 *  CLK        |_| |_| |_| |_| |_| |_| |_| |_|
 *            ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI -----X___X___X___X___X___X___X___X___XXX----
 * MISO        7   6   5   4   3   2   1   0
 * 
 * 
 *      _____      Mode 3 : CPOL=1  CPHA=1      ____
 *   CS      |_________________________________|
 *      _______   _   _   _   _   _   _   _   ______
 *  CLK        |_| |_| |_| |_| |_| |_| |_| |_|
 *              ___ ___ ___ ___ ___ ___ ___ ___
 * MOSI -----XXX___X___X___X___X___X___X___X___X----
 * MISO          7   6   5   4   3   2   1   0
 * 
 *************************************************************************
 * 
 * Propeller 2 I/O has no externally clocked registers.  Everything is
 * synchronous to the internal sysclock, ie:  It can't do a true
 * synchronous serial port.
 * 
 * This is emulated by using logic compares within each smartpin.  This
 * results in notable CLK input to MISO data output lag, measured in sysclocks.
 * Also, smartpins are located in the middle of the IC for bussing reasons.
 * This means they also require staged signals to reach their respective pins.
 * Nominally the figure is internally 3 sysclock ticks:
 * One for input stage
 * One for output stage
 * One for smartpin maths
 * Plus the asynchronous sampling of CLK adds a partial extra tick
 * 
 * Here's what that looks like at lower speeds:
 * 
 *      _____     Prop2 Lag Effect Mode 0       ____
 *   CS      |_________________________________|
 *              _   _   _   _   _   _   _   _
 *  CLK _______| |_| |_| |_| |_| |_| |_| |_| |______
 *          ______ ___ ___ ___ ___ ___ ___ ___
 * MISO ___X______X___X___X___X___X___X___X___XXX___
 *              7   6   5   4   3   2   1   0
 * 
 * Which would work for mode 1 just as well as mode 0.
 * 
 * Instead of living with the nominal lagging effect, which limits max speed,
 * we can advance the data phase just by inverting CLK at the shifter.  Here's
 * what happens in the Prop2 when applying this inversion:
 * 
 *      _____         Effective Mode 0          ____
 *   CS      |_________________________________|
 *              _   _   _   _   _   _   _   _
 *  CLK _______| |_| |_| |_| |_| |_| |_| |_| |______
 *          ____ ___ ___ ___ ___ ___ ___ ___
 * MISO ___X____X___X___X___X___X___X___X___XXXXX___
 *            7   6   5   4   3   2   1   0
 * 
 * 
 *      _____         Effective Mode 1          ____
 *   CS      |_________________________________|
 *              _   _   _   _   _   _   _   _
 *  CLK _______| |_| |_| |_| |_| |_| |_| |_| |______
 *          ______ ___ ___ ___ ___ ___ ___ ___
 * MISO ___X______X___X___X___X___X___X___X___XXX___
 *              7   6   5   4   3   2   1   0
 * 
 * 
 *      _____         Effective Mode 2          ____
 *   CS      |_________________________________|
 *      _______   _   _   _   _   _   _   _   ______
 *  CLK        |_| |_| |_| |_| |_| |_| |_| |_|
 *          ____ ___ ___ ___ ___ ___ ___ ___
 * MISO ___X____X___X___X___X___X___X___X___XXXXX___
 *            7   6   5   4   3   2   1   0
 * 
 * 
 *      _____         Effective Mode 3          ____
 *   CS      |_________________________________|
 *      _______   _   _   _   _   _   _   _   ______
 *  CLK        |_| |_| |_| |_| |_| |_| |_| |_|
 *          ______ ___ ___ ___ ___ ___ ___ ___
 * MISO ___X______X___X___X___X___X___X___X___XXX___
 *              7   6   5   4   3   2   1   0
 * 
 * NOTE:  Even with this applied the max data rate is still limited to
 *        sysclock/5.
 * 
 *************************************************************************
 * 
 * As a minimum, SPI requires just half clock cycle from CS low to first clock
 *      ____ <------>
 *   CS     |________________________ ......
 *                   _______
 *  CLK ____________|       |_______| ......
 * 
 */


static size_t  spi_shift_out( uint8_t *addr )
{
    size_t  count = 2;   // +2 when buffer data moves to the shifter

    __asm volatile {  // no optimising and enforces Fcache use - Needed to free up the FIFO
		rdfast	#0, addr    // setup the hubRAM FIFO hardware for sequencial reads
		neg	pa, #CLK & 0x20   wc    // SETPAT port select:  C=0 for INA, C=1 for INB
		modz	15   wz    // set Z flag:  SETPAT equality (INx == S) event
		setpat	mask1, mask1    // Setup CLK glitch event:  When CS high and CLK high
		setse1	#2<<6 | CS    // Setup CS falling event:  1=rising-edge, 2=falling-edge

		rflong	pa    // read four bytes from FIFO (hubRAM)
		movbyts	pa, #0b00_01_10_11    // 32-bit byte-swap to suit big-endian shifted out from lsb of register
		rev	pa    // 32-bit bit-swap to suit big-endian shifted out from lsb of register
.retry
		dirl	#MISO    // disable/reset SPI TX smartpin, pin is driven low
		wypin	pa, #MISO    // place first two bytes directly in smartpin's shift register
		dirh	#MISO    // enable SPI TX smartpin, first bit is presented to pin
.wait
		jpat	#.retry    // CLK toggled while CS still high, treat as a glitch
		jnse1	#.wait    // wait while CS is high

		setse1	#1<<6 | CS    // Setup CS rising event:  1=rising-edge, 2=falling-edge
		movbyts	pa, #0b01_00_11_10  // move second lot down
		wypin	pa, #MISO    // place second lot in buffer register

.loop
		testp	#MISO   wc    // check for empty smartpin buffer register, 32 bits shifted out
	if_c	rfword	pa    // read two bytes from FIFO (hubRAM)
	if_c	movbyts	pa, #0x1b    // 32-bit byte-swap to suit big-endian shifted out from lsb of register
	if_c	rev	pa    // 32-bit bit-swap to suit big-endian shifted out from lsb of register
	if_c	wypin	pa, #MISO    // place another two bytes in buffer register
	if_c	add	count, #2
	_ret_	jnse1	#.loop    // break loop when CS high

mask1		long	(1<<(CS & 0x1f)) | (1<<(CLK & 0x1f))   // CS/CLK pin mask for SETPAT
    }  // Exiting from Fcache/Cogexec triggers an implicit RDFAST by the branch to hubexec

    return  count;
}



static void  filldat( uint8_t *addr, size_t len )
{
    for( size_t i = 0; i < len; i++ )    // bytes
        addr[i] = i;
}


void main(void)
{
    uint32_t  m, x, ret;
    uint8_t  txdata[500];

    _waitms(200);
    printf( "\n   clkfreq = %d   clkmode = 0x%x\n", _clockfreq(), _clockmode() );

    _wrpin(CS, P_HIGH_15K | P_LOW_FLOAT | P_SCHMITT_A);  // do this first
    _drvh(CS);  // 15 kR pull-up resistor

    _wrpin(CLK, (CPOL ? 0 : P_INVERT_A) | P_SCHMITT_A);  // This invert doesn't apply to other smartA/B muxes

    m = P_SYNC_TX | P_SCHMITT_A | P_OE;
    m |= ((CLK - MISO) & 0b0111) << 24;  // smartB (clock pin) input select
    m |= (CPOL ^ !CPHA) ? 0 : P_INVERT_B;  // SPI Clock Mode (see above documentation)
    x = 0<<5 | (32-1);  // continuous buffer mode and 32-bit shifter and buffer
    _pinstart(MISO, m, x, 0);


//    memset( addr, 0, sizeof(txdata) );
    filldat( txdata, sizeof(txdata) );
    x = 0;

    while(1)
    {
        txdata[0] = x;
        ret = spi_shift_out( txdata );
        printf( " BlockID = %d    Byte Count = %d\n", x, ret );
        x = (x + 1) % 20;
    }
}
